#include "common.h"
#include "http.h"

#define IS_HTTP_11	(opts.flags & OPT_HTTP_11)

static int http_parse_response(char* buff, http_response* resp);

int
http_parse_url(const char* loc, http_url* url)
{
	char *lloc, *ptr, *ptr2;
	int i, urllen;

	urllen = strlen(loc);
	if (strlen(loc) < 8) return -1;		/* malformed url */

	lloc = (char*) my_alloc(strlen(loc)+1);
	strcpy(lloc, loc);
	
	for (i = 0; i < 4; i++) 
		lloc[i] = (char)tolower(lloc[i]);
	if (strncmp(lloc, "http://", 7)) 
	{
		return -1;
	}
	strcpy(url->plain_url, lloc);

	ptr2 = ptr = lloc + 7;
	while (*ptr2 != '/' && *ptr2 != ':' && *ptr2 != 0) 
	{
		if (++ptr2 - ptr > MAX_SERVER_NAME_LEN) return -1;
	}

	memcpy(url->server, ptr, ptr2-ptr);
	url->server[ptr2-ptr] = 0;
	
	if (*ptr2 == ':') 
	{
		char buff[6];
		ptr = ptr2+1;
		while (isdigit(*++ptr2));
		if (ptr2 - ptr > 5) return -1;
		memcpy(buff, ptr, ptr2-ptr);
		buff[ptr2-ptr] = 0;
		url->port = atoi(buff);
		if (url->port > 65536) return -1;
	}
	else url->port = 80;
	if (*ptr2 && *ptr2 != '/') return 0;
	if (*ptr2 == '/' && *(ptr2+1)) 
	{
		strcpy(url->location, ptr2);
		return 0;
	}
	
	strcpy(url->location, "/");
	
	return 0;
}

int
http_init_request(http_request* req, int meth)
{
	req->url 		= NULL;
	req->method		= meth;
	req->user_agent[0] = 0;
	req->referer[0]	   = 0;
	req->from[0]	   = 0;
	req->if_modified[0]= 0;
	req->auth[0]	   = 0;
	req->body		   = NULL;
	req->body_len      = 0;
	return 0;
}

char*
http_create_request(http_request* req, int* len)
{
	char *req_ptr, *nptr, *mtype;
	int  req_len;
	char auth[256], auth_enc[256];
	int al;

	req_len = sizeof(http_request) + 256 + req->body_len;
	req_ptr = (char*) my_alloc(req_len);
	nptr    = req_ptr;

	switch (req->method) 
	{
		case HTTP_METHOD_GET:
			mtype = "GET";
			break;
		case HTTP_METHOD_POST:
			mtype = "POST";
			break;
		case HTTP_METHOD_HEAD:
			mtype = "HEAD";
			break;
		default:
			free(req_ptr);
			return NULL;
	}
	sprintf(nptr, "%s %s HTTP/1.%d\n", mtype, 
			req->url->proxy_server[0] ? req->url->plain_url : req->url->location,
			IS_HTTP_11 ? 1 : 0);
	nptr = req_ptr + strlen(req_ptr);
	if (req->user_agent[0]) 
	{
		sprintf(nptr, "User-Agent: %s\n", req->user_agent);
		nptr += strlen(nptr);
	}
	if (req->url->username[0]) 
	{
		sprintf(auth, "%s:%s", req->url->username, req->url->password);
		al = base64enc(auth, strlen(auth), auth_enc);
		sprintf(nptr, "Authorization: Basic %.*s\n", al, auth_enc);
		nptr += strlen(nptr);
	}
	if (IS_HTTP_11)
	{
		sprintf(nptr, "Connection: close\nHost: %s\n", req->url->server);
		nptr += strlen(nptr);
	}
	/* other headers to come here... */
	*nptr++ = '\n'; *nptr++ = 0;
	if (req->body && req->body_len) 
	{
		memcpy(--nptr, req->body, req->body_len);
		nptr += req->body_len;
	}
	*len = (int)(nptr - req_ptr);
	return req_ptr;
}

int 
http_send(http_request* req, http_response* response)
{
	static int fc = 1;
	int fd, rc;
	char *bptr;
	int isize = 0;
	int request_len;
	struct sockaddr_in server;
	struct hostent *he;
	char  in_buff[16000];
	char *out_req;
	unsigned bytes_read = 0, bytes_wrote = 0;

	memset((char*)&server, 0, sizeof(server));
	server.sin_family	= AF_INET;
	
	if (req->url->proxy_server[0])
	{
		server.sin_port		= htons(req->url->proxy_port);
		he = gethostbyname(req->url->proxy_server);
		if (!he) 
		{
			return HG_ERR_UNRESOLVED;
		}
		DEBUGPR(("ProxyIP:  %s\n", inet_ntoa(*((struct in_addr*)(he->h_addr)))));
#ifdef	WITH_LOG
		if (fc)
			write_log(log_name, wl_debug, "Proxy IP:  %s", 
					inet_ntoa(*((struct in_addr*)(he->h_addr))));
#endif
	}
	else
	{
		server.sin_port		= htons(req->url->port);
		he = gethostbyname(req->url->server);
		if (!he) 
		{
			return HG_ERR_UNRESOLVED;
		}
		DEBUGPR(("ServerIP:  %s\n", inet_ntoa(*((struct in_addr*)(he->h_addr)))));
#ifdef	WITH_LOG
		if (fc) 
			write_log(log_name, wl_debug, "Server IP:  %s", 
				inet_ntoa(*((struct in_addr*)(he->h_addr))));
#endif
	}
	
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0) 
	{
		return HG_ERR_SYS;	/* system error */
	}
	memcpy((char*)&server.sin_addr.s_addr, he->h_addr, he->h_length);
	alarm(HTTP_CONNECT_TIMEOUT);
	if (connect(fd, (struct sockaddr*)&server, sizeof(server)) < 0) 
	{
		if (errno == EINTR) /* this will never should happen on Linux ? */
		{
			write_log(log_name, wl_error, "connection timed out");
		}
		return HG_ERR_SYS;
	}
	alarm(0);
	
	out_req = http_create_request(req, &request_len);
	if (!out_req) 
	{
		printf("error in http request data\n");
		return HG_ERR_HTTP;
	}

	DEBUGPR(("outgoing request:<<\n%s>>\n", out_req));
	
	if ((rc = write(fd, out_req, request_len)) < 0) 
	{
		return HG_ERR_SYS;
	}
	bytes_wrote = rc;
	
	bptr = in_buff;
	while ((rc = read(fd, bptr + isize, MAX_BUFF_LEN)) > 0) 
	{
		/* DEBUGPR((in_buff)); */
		isize += rc;
	}
	DEBUGPR(("received %d bytes from server", isize));
	bytes_read = isize;
	if (http_parse_response(in_buff, response) < 0) 
	{
		fprintf(stderr, "error: wrong data in header\n");
		return HG_ERR_HTTP;
	}
	DEBUGPR(("http version: %d.%d", response->http_ver_major, response->http_ver_minor));
	DEBUGPR(("status code:  %d", response->status_code));
	DEBUGPR(("authentication:  %s", response->www_auth));
	
	close(fd);
	if (!fc)
	{
		char buff[128];
		sprintf(buff, "t%d %d", bytes_read, bytes_wrote);
		send_to_main(buff);
	}
	fc = 0;

	if (opts.flags & OPT_SAVE_RESPONSE || 
	   ((opts.flags & OPT_SAVE_RESPONSE_200) && response->status_code == 200))
	{
		FILE* f;
		if ((f = fopen(opts.saved_response, "w")) == NULL)
		{
			write_log(log_name, wl_error, "couldnt open file %s",
					opts.saved_response);
		}
		else
		{
			fwrite(in_buff, isize, 1, f);
			fclose(f);	
		}
	}
	
	return 0;
}

int 
http_parse_response(char* buff, http_response* resp)
{
	char *ptr, *ptr2;
	char lbuff[MAX_BUFF_LEN];
	char dummy[MAX_HTTP_HEADER_LEN];
	ptr = buff;
	memset(resp, 0, sizeof(http_response));
	resp->content	= NULL; 	/* just in case :-) */
	if (strncmp(ptr, "HTTP/", 5))
		return -1;
	while (1) 
	{
		ptr2 = strchr(ptr, (int)'\n');
		if (!ptr2) break;
		if (*(ptr2+1) == '\n')
		{
			/* 
			 * content begins 
			 */
			ptr2 += 2;
			if (resp->content_length)
			{
				resp->content = (char*)my_alloc(resp->content_length);
				memcpy(resp->content, ptr2, resp->content_length);	
			}
			else
			{
				int cl;
				if (resp->content_type != HTTP_CONTENT_HTML)
				{
					write_log(log_name, wl_error, 
							  "cannot determine content length for the response");
					return -1;
				}
				cl = strnlen(ptr2, MAX_TXT_CONTENT_LEN);
				resp->content = (char*)my_alloc(cl);
				memcpy(resp->content, ptr2, cl);
				resp->content_length = cl;
			}
			break;
		}
		memcpy(lbuff, ptr, ptr2-ptr);
		lbuff[ptr2-ptr] = 0;
		if (!strncmp(lbuff, "HTTP/", 5)) 
		{
			sscanf(lbuff, "HTTP/%d.%d %d", 
					&resp->http_ver_major, 
					&resp->http_ver_minor,
					&resp->status_code);
			strcpy(resp->reason_phrase, lbuff+13);
			resp->reason_phrase[strlen(resp->reason_phrase)-1] = 0;
		}
		else if (!strncasecmp(lbuff, "Server:", 7)) 
		{
			sscanf(lbuff, "Server: %s", resp->server);
		}
		else if (!strncasecmp(lbuff, "WWW-Authenticate:", 10)) 
		{
			sscanf(lbuff, "WWW-Authenticate: %s", resp->www_auth);
		}
		else if (!strncasecmp(lbuff, "Location:", 8)) 
		{
			sscanf(lbuff, "Location: %s", resp->location);
		}
		else if (!strncasecmp(lbuff, "Content-Length:", 10)) 
		{
			sscanf(lbuff, "Content-Length: %ld", &resp->content_length);
		}
		else if (!strncasecmp(lbuff, "Content-Type:", 10)) 
		{
			/* 
			 * it would be nice to add more content-types
			 * but there is no practical use for them within WPC
			 */
			char ctbuff[MAX_HTTP_HEADER_LEN];
			sscanf(lbuff, "%13c %s", dummy, ctbuff);
			if (!strncasecmp(ctbuff, HTTP_CONTENT_HTML_STR, 8))
				resp->content_type	= HTTP_CONTENT_HTML;
			else
				resp->content_type	= HTTP_CONTENT_UNKN;
		}
		ptr = ptr2 + 1;
		if (ptr >= buff + strlen(buff)) break;
	}
	
	/* quick hack, gotta handle -M more gracefuly :-) */
	if (opts.reject_cond_code)
	{
		if (opts.match_string[0] && resp->status_code == opts.reject_cond_code 
			&& strstr(buff, opts.match_string))
			resp->status_code = 401;
	}
	else
	{
		if (opts.match_string[0] && resp->status_code == 200 
			&& strstr(buff, opts.match_string))
			resp->status_code = 401;
	}
	
	return 0;	
}

