#include "common.h"

struct options	opts;

void usage();

int
main(int argc, char** argv)
{
	int op;
	int flag_m = 0, flag_t = 0;
	char optbuff[128];

	fprintf(stderr, "Web Password Checker, v%s\n", VER); 
	
	if (argc < 2) 
	{
		usage();
		exit(EXIT_FAILURE);
	}

	memset(&opts, 0, sizeof(opts));
	opts.proxy_port		 = 8080;
	opts.flags			 = 0;
	opts.max_proc		 = 5;
	opts.transf_value	 = 0;
	opts.req_delay		 = 2;
	opts.ignore_codes	 = NULL;
	opts.flags 			|= OPT_NOTRANSF;
	opts.flags 			|= OPT_HTTP_11;

	while ((op = getopt(argc, argv, 
			"u:p:f:s:d:e:cglLt:m:y:q:H:4:3:5:2:M:a:A:r:")) != EOF) 
	{
		if (op == '?')
			exit(EXIT_FAILURE);
		if (op == ':')
			exit(EXIT_FAILURE);

		switch (op)
		{
			case 'u':
				opts.flags |= OPT_DEF_USER;
				strcpy(opts.def_username, optarg);
				break;
				
			case 'p':
				opts.flags |= OPT_DEF_PASS;
				strcpy(opts.def_password, optarg);
				break;
				
			case 'f':
				opts.max_proc = atoi(optarg);
				if (!opts.max_proc || opts.max_proc > MAX_PROC) 
				{
					fprintf(stderr, "wrong number of processes specified\n");
					exit(EXIT_FAILURE);
				}
				break;

			case 's':
				{
					char* cptr = strchr(optarg, ':');
					if (!cptr)
					{
						strcpy(opts.proxy_server, optarg);
					}
					else
					{
						memcpy(opts.proxy_server, optarg, cptr-optarg);
						opts.proxy_server[cptr-optarg] = 0;
						cptr++;
						if (*cptr)
						{
							opts.proxy_port = atoi(cptr);
							if (!opts.proxy_port)
							{
								fprintf(stderr, "wrong proxy settings\n");
								exit(EXIT_FAILURE);
							}
						}
					}
				}
				break;

			case 'a':
				strcpy(opts.saved_response, optarg);
				opts.flags |= OPT_SAVE_RESPONSE;
				break;
				
			case 'A':
				strcpy(opts.saved_response, optarg);
				opts.flags |= OPT_SAVE_RESPONSE_200;
				break;
				
			case 'd':
				strcpy(opts.dict_list_file, optarg);
				break;

			case 'e':
				opts.flags |= OPT_DEF_PASSES;
				strcpy(opts.def_pass_dict, optarg);
				break;
				
			case '2':
			case '3':
			case '4':
			case '5':
				if (!opts.ignore_codes)
				{
					mylist_init(&opts.ignore_codes);
				}
				optbuff[0] = op; optbuff[1] = '\0';
				strcat(optbuff, optarg);
				if (!atoi(optbuff) || strlen(optbuff) != 3)
				{
					fprintf(stderr, "%s: wrong response code given\n", argv[0]);
					exit(EXIT_FAILURE);
				}	
				mylist_add(opts.ignore_codes, optbuff, 4);
				opts.flags |= OPT_IGNORECODES;
				break;
				
			case 'c':
				opts.flags |= OPT_DICT_COMBO;
				break;

			case 'g':
				opts.flags |= OPT_HTTP_GET;
				break;

			case 'l':
				opts.flags |= OPT_LOG;
				break;
				
			case 'L':
				opts.flags |= OPT_LOG;
				opts.flags |= OPT_LOG_NOCOMB;
				break;
				
			case 't':
				flag_t++;
				{
					int ti = atoi(optarg);
					if (ti < 1 || ti > 9)
					{
						fprintf(stderr, "%s: wrong transformation value specified\n", argv[0]);
						exit(EXIT_FAILURE);
					}
					opts.flags &= ~OPT_NOTRANSF;
					if (ti < 9)
					{
						opts.flags |= OPT_TRANSF_LESS;
					}
				}
				break;

			case 'm':
				flag_m++;
				if (optarg[0] == 'U' || optarg[0] == 'u')
					opts.flags |= OPT_DEF_USER;
				else
				if (optarg[0] == 'P' || optarg[0] == 'p')
					opts.flags |= OPT_DEF_PASS;
				else
				{
					fprintf(stderr, "%s: only 'u' or 'p' allowed with -m\n", argv[0]);
					exit(EXIT_FAILURE);
				}
				break;

			case 'M':
				strcpy(opts.match_string, optarg);
				break;

			case 'y':
				opts.req_delay = atoi(optarg);
				if (opts.req_delay < 1 || opts.req_delay > 100000)
				{
					fprintf(stderr, "%s: wrong value of delay\n", argv[0]);
					exit(EXIT_FAILURE);
				}
				break;
				
			case 'q':
				strcpy(opts.quick_dict, optarg);
				break;
				
			case 'r':
				opts.reject_cond_code = atoi(optarg);
				if (opts.reject_cond_code < 200 || opts.reject_cond_code > 303)
				{
					fprintf(stderr, "%s: wrong code given with -r\n", argv[0]);
					exit(EXIT_FAILURE);
				}
				break;
				
			case 'H':
				if (atoi(optarg) == 10) opts.flags &= ~OPT_HTTP_11;
				else if (atoi(optarg) != 11) 
				{
					fprintf(stderr, "%s: wrong http version specified\n", argv[0]);
					exit(EXIT_FAILURE);
				}
				break;

			default:
				fprintf(stderr, "%s: unknown flag '%c'\n", argv[0], op);
				exit(EXIT_FAILURE);
		}

	}

	/* make -m p default option if no combo */
	if (!(opts.flags & OPT_DICT_COMBO) && !(opts.flags & OPT_DEF_USER))
	{
		opts.flags |= OPT_DEF_PASS;
	}
	/* -m makes sense only with transformations */
	if (flag_m && !flag_t)
	{
		fprintf(stderr, "%s: -t must be specified with -m\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	/* -r only with -M */
	if (opts.reject_cond_code && !opts.match_string[0])
	{
		fprintf(stderr, "%s: -r must be specified with -M\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	if (!argv[optind])
	{
		usage();
		exit(EXIT_FAILURE);
	}
	
	write_log(log_name, wl_info, "started");
	printf("--- %s\n", argv[optind]);
	process(argv[optind]);
	write_log(log_name, wl_info, "finished");
	
	return EXIT_SUCCESS;
}

void 
usage()
{
	char msg[] = "\
usage: wpc <options> <url>\n\
options:\n\
    -s<proxy>[:<port>]  specify proxy server\n\
    -c                  assume dictionaries are combos\n\
    -e<dictionary>      apply passwords from <dictionary> to each username\n\
    -d<dict_list>       take dictionaries from <dict_list> instead of default\n\
                        file 'dicts'\n\
    -q<dictionary>      use only one dictionary specified as parameter insted of\n\
                        dictionaries specified in file 'dicts'\n\
    -t<level>           use word transformations, level=9 - maximal set,\n\
                        level=1 - minimal set (currently only 1 & 9 are valid)\n\
    -m<param>           apply word transormations either to username param=u\n\
                        or to password param=p (default), must be used with -t\n\
    -g                  use http method GET (default is HEAD)\n\
    -<nnn>              ignore response code <nnn> (multiple codes ok)\n\
    -M<string>          assume the pass is valid only if the http 200 reply\n\
                        doesn't match <string> (use with -g option)\n\
    -f<count>           use <count> processes simultaneously for sending \n\
                        requests (default is 5)\n\
    -y<delay>           sleep for <delay> miliseconds between each http request\n\
    -H<version>         use http <version>, values 11 (default) or 10 allowed
    -l                  turn logging on\n\
    -L                  the same as -l, but do not log user/pass combinations\n\
    -u<username>        use default username\n\
    -p<password>        use default password\n\
    -a<filename>        save response into file <filename>\n\
    -A<filename>        the same as -a, but save only positive responses\n\
                        (useful when breaking \"tricky\" protections)\n\
    -r<code>            ignore specific http response <code> only if the\n\
                        response contains the string given with -M\n\
";
	fprintf(stderr, msg);
}
