#include "common.h"
#include "http.h"
#include "words.h"
#include "transf.h"

/* constants more or less */
const int max_http_errors    = 20;
const int max_maybes         = 5;
const int max_valids		 = 10;
const  char def_user_agent[] = "(null)";
char   log_name[]            = "wpc.log";
char   history_name[]        = "wpc-history.txt";

extern dictionary* def_passes;
int	   totalwords = 0;
pid_t  main_pid;
int	   m_id;
char   msgbuff[256];
struct msgbuf* msgb;
int    termcolumns = 80;
int	   cracks = 0;
int    tries  = 0, maybes = 0;
int    bytes_in = 0, bytes_out = 0;
time_t start_time;
static int exit_by_alarm;

static http_request request;
static http_response resp;

static int check_auth(char *addr, const char* un, const char *pw);
static int check_loop(char *addr);
static int try_un_pw(char *addr, char *un, char *pw);
static int is_code_ignored(int c);

void  sighandler(int);
int   handle_kids();
void  show_report();
void  cleanup();

int 
process(char* addr)
{
	int rc, i, j;
	int dict_words_proc;
	pid_t pid[MAX_PROC];

	signal(SIGALRM, sighandler);
	signal(SIGINT, sighandler);
	main_pid = getpid();
	
	exit_by_alarm = 1;
	rc = check_auth(addr, NULL, NULL);
	exit_by_alarm = 0;
	if (rc == 200) 
	{
		printf("%s is not password protected\n", addr);
		return 0;
	}
	else if (rc == 407)
	{
		printf("proxy is password protected\n");
		return 0;
	}
	else if (rc == 301 || rc == 302)
	{
		printf("%s redirects to %s\n", addr, resp.location);
		return 0;
	}
	else if (rc != 401 && rc != 403) 
	{
		printf("%s returned %d\n", addr, rc);
		return 0;
	}

	write_log(log_name, wl_info, "gonna check %s", addr);
	
	if ((dcount = load_dictionaries(dicts)) < 0) 
	{
		fprintf(stderr, "error: cant load dictionaries\n");
		return -1;
	}
	for (i = 0; i <= dcount; i++) 
	{
		DEBUGPR(("dict: %s, words: %d", dicts[i]->path, dicts[i]->word_count));
		write_log(log_name, wl_info, "dict: %s, words: %d", dicts[i]->path, dicts[i]->word_count);
	}

	init_transf();

	m_id = msgget(IPC_PRIVATE, S_IRWXU);
	if (m_id < 0) 
	{
		fprintf(stderr, "msgget: %s\n", strerror(errno));
		fatal("couldnt create message queue");
	}
	msgb = (struct msgbuf*) malloc(128);

	printf("\n");	
	start_time = time(NULL);
	write_history(chomp(asctime(localtime(&start_time))));
	write_history("--- %s", addr);
	
	for (i = 0; i < opts.max_proc; i++) 
	{
		int di;
		for (di = 0, j = 0; dicts[j] && j <= dcount; di++, j++) 
		{
			if (dicts[j]->word_count < (opts.max_proc << 1)) 
			{
				if (!i) 
				{
					word_dist[di].dict_index = j;
					word_dist[di].start = 1;
					word_dist[di].end   = dicts[j]->word_count;
				}
				else 
				{
					di--;
					continue;
				}
			}
			else 
			{
				dict_words_proc = dicts[j]->word_count / opts.max_proc;
				word_dist[di].dict_index = j;
				word_dist[di].start = i * dict_words_proc + 1;
				word_dist[di].end   = word_dist[di].start + dict_words_proc- 1;
				if (i == opts.max_proc - 1) 
				{
					word_dist[di].end += dicts[j]->word_count % dict_words_proc;
				}
			}
		}
		word_dist[di].dict_index = -1;
		rc = fork();
		if (rc < 0) 
		{
			fprintf(stderr, "fork failed, %s\n", strerror(errno));
			return -1;
		}
		if (rc == 0) 
		{
			exit(check_loop(addr));
		}
		pid[i] = rc;
		
	}

	atexit(cleanup);
	atexit(show_report);
	handle_kids();
	
	return 0;
	
}

int handle_kids()
{
	int pid_status, rc;
	int i = opts.max_proc;
	char outpbuff[128];
	int wordtries = 0;

	/* wait and look at the childs */
	while (i)
	{
		rc = waitpid(-1, &pid_status, WNOHANG);
		if (rc)
		{
			write_log(log_name, wl_debug, "child %d exited", rc);
			i--;
			continue;
		}
		else
		{
			milisleep(5);
			rc = msgrcv(m_id, msgb, 128, 0, IPC_NOWAIT);
			if (rc < 0 && (errno == EAGAIN || errno == ENOMSG)) 
				continue;
			else if (rc < 0)
			{
				fprintf(stderr, "\r                                            ");
				fprintf(stderr, "\n\nmsgrcv failed: %d %s\n\n", errno, strerror(errno));
				return -1;
			}
			else
			{
				char cmd;
				int  bi, bo;
				
				cmd = msgb->mtext[0];
				switch (cmd)
				{
					case '+':
						cracks++;
						if (cracks > max_valids)
						{
							write_log(log_name, wl_error, "maximum count of valid"
									 " passes reached, exiting");
							fprintf(stderr, "error: too many quasi-valid passes\n");
							return -1;
						}
						fprintf(stderr, "\r                                            ");
						fprintf(stderr, "\n*** valid pass: %s\n\n", msgb->mtext+1);
						write_history("    valid: %s", msgb->mtext+1);
						break;
						
					case '?':
						maybes++;
						if (maybes > max_maybes)
						{
							write_log(log_name, wl_error, "maximum count of possibly valid"
									 " passes reached, exiting");
							fprintf(stderr, "error: too many quasi-valid passes\n");
							return -1;
						}
						fprintf(stderr, "\r                                            ");
						fprintf(stderr, "\n*** may be valid: %s\n\n", msgb->mtext+1);
						write_history("    may be valid: %s", msgb->mtext+1);
						break;
						
					case '-':
						tries++;
						sprintf(outpbuff, "\r[%.5d/% 3d%%] trying: %s                    \r", 
								tries, wordtries*100/totalwords,
								msgb->mtext+1);
						fprintf(stderr, "%.*s", termcolumns, outpbuff);
						break;
						
					case 'w':
						wordtries++;
						break;

					case 't':
						sscanf(msgb->mtext+1, "%d %d", &bi, &bo);
						bytes_in  += bi;
						bytes_out += bo;
						break;
						
					default:
						break;
				}
			}
			
		}
	}
	return 0;
}

void cleanup()
{
	int rc;
	rc = msgctl(m_id, IPC_RMID, NULL);
	if (rc < 0)
	{
		fprintf(stderr, "can't clear message queue, %s\n", strerror(errno));
	}
}

void 
show_report()
{
	char tbuff[128];
	char lbuff[128];
	time_t elt, elh, elm, els;
	if (main_pid != getpid() || !tries) 
		return;
	elt = time(NULL) - start_time;
	elh = elt/3600;
	elm = elt>3600 ? (elt%3600)/60 : elt/60;
	els = elt>60 ? (elt%60) : elt;
	sprintf(tbuff, "%.2ld:%.2ld:%.2ld", elh, elm, els); 
	printf("\n--------------------------------\n");
	sprintf(lbuff, "Elapsed time:  %s", tbuff);
	write_log(log_name, wl_info, lbuff);
	puts(lbuff);
	sprintf(lbuff, "Bytes in/out:  %d/%d", bytes_in, bytes_out);
	write_log(log_name, wl_info, lbuff);
	puts(lbuff);
	sprintf(lbuff, "Attempts:      %d", tries);
	write_log(log_name, wl_info, lbuff);
	puts(lbuff);
	sprintf(lbuff, "Cracks:        %d", cracks);
	write_log(log_name, wl_info, lbuff);
	puts(lbuff);
	write_history("");
}


int 
check_loop(char *addr)
{
	char guess_username[MAX_USERNAME_LEN];
	char guess_password[MAX_PASSWORD_LEN];
	static char saved_username[MAX_PASSWORD_LEN];
	int rc;

#if defined(DEBUG)
	int i;
	for (i = 0; word_dist[i].dict_index != -1; i++) 
	{
		printf("*** [%d] %d: %d->%d\n", getpid(), i+1, word_dist[i].start, word_dist[i].end);
	}
#endif

	DEBUGPR(("before in loop"));
	while (get_un_pw(guess_username, guess_password) == 0) 
	{
		DEBUGPR(("in loop"));
			
		rc = try_un_pw(addr, guess_username, guess_password);
		if (rc == 0) break;
		if (rc == -2) goto finish;
		if ((opts.flags & OPT_DEF_PASSES) && strcmp(guess_username, saved_username))
		{
			int start = 1;
			char *pass;
			while (pass = get_word(&start, def_passes))
			{
				rc = try_un_pw(addr, guess_username, pass);
				if (rc == 0) goto finish;
			}
			strcpy(saved_username, guess_username);
		}
	}
finish:
	return EXIT_SUCCESS;
}

int
try_un_pw(char *addr, char *un, char *pw)
{
	static unsigned http_errors;
	int rc;
	
	sprintf(msgbuff, "-%s/%s", un, pw);
	if (send_to_main(msgbuff) < 0)
	{
		fatal("can not connect to main process");
	}
	fflush(stdout);
	if (!(opts.flags & OPT_LOG_NOCOMB))
		write_log(log_name, wl_info, "trying %s/%s", un, pw);
	rc = check_auth(addr, un, pw);
	if (rc == 200)	
	{
		sprintf(msgbuff, "+%s:%s", un, pw);
		send_to_main(msgbuff);
		write_log(log_name, wl_info, "*** valid pass: %s:%s", un, pw);
	}
	if (!is_code_ignored(rc) && ((rc >= 300 && rc <= 304) || rc == 404))
	{
		sprintf(msgbuff, "?%s:%s", un, pw);
		send_to_main(msgbuff);
		write_log(log_name, wl_info, "*** pass may be valid: %s:%s", un, pw);
		write_log(log_name, wl_info, "(redirect to %s", resp.location);
	}
	if (rc != 200 && rc != 401 && !is_code_ignored(rc))  
			http_errors++;
	if (http_errors > max_http_errors)
	{
		fprintf(stderr, "\nerror: maximum error count reached\n");
		return -2;
	}
	if (rc == 502 || rc == 503)
	{
		write_log(log_name, wl_info, "error 502/503 occured, sleeping for 10 seconds");
		sleep(10);
	}
	milisleep(opts.req_delay);
	return -1;
}

int 
check_auth(char *addr, const char* un, const char *pw)
{
	static int fc = 1;
	static http_url url;
	static char *paddr = NULL;
	int rc;
	
	if (fc || paddr != addr)
	{
		if (http_parse_url(addr, &url) < 0) 
		{
			fatal("malformed url");
		}
		if (opts.proxy_server[0])
		{
			strcpy(url.proxy_server, opts.proxy_server);
			url.proxy_port = opts.proxy_port;
		}
		http_init_request(&request, (opts.flags & OPT_HTTP_GET) ? HTTP_METHOD_GET : HTTP_METHOD_HEAD);
		request.url	= &url;
		strcpy(request.user_agent, def_user_agent);
		
		DEBUGPR(("Server:    %s", url.server));
		DEBUGPR(("Port:      %d", url.port));
		DEBUGPR(("Location:  %s", url.location));
		
		paddr = addr;
		fc = 0;
	}
	
	if (un && pw) 
	{
		strcpy(url.username, un);
		strcpy(url.password, pw);
	}
	else 
	{
		url.username[0] = 0;
		url.password[0] = 0;
	}
	
	if ((rc = http_send(&request, &resp)) < 0) 
	{
		switch (rc) 
		{
			case HG_ERR_SYS:
				fprintf(stderr, "system error: %s\n", strerror(errno));
				break;
			case HG_ERR_UNRESOLVED:
				fprintf(stderr, "error: unknown host\n");
				break;
			default:
				fprintf(stderr, "unknown error occured\n");
				break;
		}
		fatal("execution aborted");
	}
	if (resp.status_code == 200) 
	{
		return resp.status_code;
	}
	if (resp.status_code != 401 && !is_code_ignored(resp.status_code))
	{
		fprintf(stderr, "Page is not protected, but some other error occured\n"
			   "(response code %d [%s])\n", resp.status_code, resp.reason_phrase);
	}
	return resp.status_code;
}

int 
is_code_ignored(int c)
{
	char *cptr;
	long len;
	if (!opts.ignore_codes)
		return 0;
	cptr = mylist_get_first(opts.ignore_codes, &len);
	do
	{
		DEBUGPR(("cptr = %p, len = %d, cptrs = %s", 
				cptr, len, cptr ? cptr : "(null)"));
		if (cptr && len == 4 && atoi(cptr) == c)
			return 1;
	}	while (cptr = mylist_get_next(opts.ignore_codes, &len));
	return 0;
}

int 
send_to_main(const char* text)
{
	struct msgbuf*	msgb;
	int rc;
	
	msgb = (struct msgbuf*) malloc(128);
	msgb->mtype = getpid();
	strcpy(msgb->mtext, text);
	rc = msgsnd(m_id, msgb, 128, 0);
	if (rc < 0)
	{
		fprintf(stderr, "msgsnd failed, %s\n", strerror(errno));
		return -1;
	}
		
	return 0;
}

void 
sighandler(int sig)
{
	write_log(log_name, wl_debug, "signal %d received", sig);
	if (sig == SIGALRM)
	{
		/* I know this is silly, I will make normal timeouts soon */
		if (exit_by_alarm)
		{
			fprintf(stderr, "\nerror: connection timed out\n");
			write_log(log_name, wl_debug, "connection timed out, exited");
			exit(EXIT_FAILURE);
		}
		signal(SIGALRM, sighandler);
		return;
	}
	if (sig == SIGINT)
	{
		if (getpid() == main_pid) 
		{
			fprintf(stderr, "\n*interrupted*\n");
			write_log(log_name, wl_info, "program interrupted");
		}
		exit(EXIT_FAILURE);
	}
}
