#include "common.h"
#include "transf.h"

trans_func trans_flist[MAX_TRANS_FUNCS] = 
{
	transf_reverse,
	transf_upper,
	transf_capitalize, 
	transf_add_digits_after,
	transf_add_digits_before,
	transf_add_common_digits_before,
	transf_add_common_digits_after,
	NULL
};

void	init_transf()
{
	int i = 0;
	if (opts.flags & OPT_TRANSF_LESS)
	{
		trans_flist[i++]	= transf_reverse;
		trans_flist[i++]	= transf_less_add_common_digits_after;
		trans_flist[i]		= NULL;
	}
}

char*
transf_reverse(char* str)
{
	static char saved[MAX_PASSWORD_LEN];
	static int fc = 1;
	char *ptr;
	int i = 0;
	if (!str[0]) return NULL; 
	if (fc) 
	{
		strcpy(saved, str);
		ptr = str;
		i = strlen(str);
		while (*ptr) 
		{
			saved[--i] = *ptr++;
		}
		fc = 0;
		return saved;
	}
	fc = 1;
	return NULL;
}

char*
transf_upper(char* str)
{
	static char saved[MAX_PASSWORD_LEN];
	static int fc = 1;
	char *ptr;

	if (!str[0]) return NULL;
	if (fc) 
	{
		ptr = saved;
		strcpy(saved, str);
		while (*str) *ptr++ = toupper(*str++); 
		fc = 0;
		return saved;
	}
	fc = 1;
	return NULL;
}

char*
transf_capitalize(char* str)
{
	static char saved[MAX_PASSWORD_LEN];
	static int fc = 1;
	char *ptr;

	if (!str[0]) return NULL;
	if (fc) 
	{
		ptr = saved;
		strcpy(saved, str);
		*str = toupper(*str); 
		fc = 0;
		return saved;
	}
	fc = 1;
	return NULL;
}

char*
transf_add_digits_after(char* str)
{
	static char saved[MAX_PASSWORD_LEN];
	const int  maxn = 9;
	static int curn = maxn;

	if (!str[0]) return NULL;
	if (curn) 
	{
		sprintf(saved, "%s%d", str, curn--);
		return saved;
	}
	else 
	{
		curn = maxn;
		saved[0] = 0;
		return NULL;
	}
}

char*
transf_add_digits_before(char* str)
{
	static char saved[MAX_PASSWORD_LEN];
	const int  maxn = 9;
	static int curn = maxn;

	if (curn) 
	{
		sprintf(saved, "%d%s", curn--, str);
		return saved;
	}
	else 
	{
		curn = maxn;
		saved[0] = 0;
		return NULL;
	}
}

char*
transf_add_common_digits_before(char* str)
{
	static char* common_digits[] = 
	{
		"12", "123", "1234", "12345",
		"54321", "4321", "321", "21",
		"11", "22", "33", "66", "69", "99",
		"111", "666", "777",
		"2000", "1999", "1998", "1997",
		"6969", "9696",
		NULL
	};
	static char saved[MAX_PASSWORD_LEN];
	static int curidx = 0;

	if (!str[0]) return NULL;
	if (!common_digits[curidx]) 
	{
		curidx = 0;
		return NULL;
	}
	sprintf(saved, "%s%s", common_digits[curidx++], str);
	return saved;
}

char*
transf_add_common_digits_after(char* str)
{
	static char* common_digits[] = 
	{
		"12", "123", "1234", "12345",
		"54321", "4321", "321", "21",
		"11", "22", "33", "66", "69", "99",
		"111", "666", "777",
		"2000", "1999", "1998", "1997",
		"6969", "9696",
		NULL
	};
	static char saved[MAX_PASSWORD_LEN];
	static int curidx = 0;
	if (!common_digits[curidx]) 
	{
		curidx = 0;
		return NULL;
	}
	sprintf(saved, "%s%s", str, common_digits[curidx++]);
	return saved;
}

char*
transf_less_add_common_digits_after(char* str)
{
	static char* common_digits[] = 
	{
		"1", "2", "69", "123",  NULL
	};
	static char saved[MAX_PASSWORD_LEN];
	static int curidx = 0;
	if (!common_digits[curidx]) 
	{
		curidx = 0;
		return NULL;
	}
	sprintf(saved, "%s%s", str, common_digits[curidx++]);
	return saved;
}
