#include "common.h"
#include "words.h"
#include "transf.h"

char def_dict_paths_file[] = "dicts";
word_distribution	word_dist[MAX_DICTS];

dictionary *dicts[MAX_DICTS];
dictionary *def_passes = NULL;
int	dcount;

static int load_dict(char*, dictionary**);

int
get_un_pw(char* un, char* pw)
{
	const  int max_dict_errors = 15;
	static int dict_errors = 0;
	static int fc = 1;
	static int wdi = 0;
	static int curr_dict = 0;
	static int curr_dict_word;
	static char saved_word[MAX_PASSWORD_LEN];
	static int  curr_trans = 0;
	static char saved_un[MAX_USERNAME_LEN];
	char *ptr;

	if (word_dist[0].dict_index < 0) return -1;

	if (fc) 
	{
		fc = 0;
		curr_dict_word = word_dist[0].start;
		DEBUGPR(("loaded %d dict(s)", dcount+1));
	}

get_next_word:

	if (!curr_trans) 
	{
		if (curr_dict == -1) 
		{
			DEBUGPR(("finish"));
			return -2;
		}
		
		ptr = get_word(&curr_dict_word, dicts[curr_dict]);
		if (!ptr) 
		{
			if (word_dist[++wdi].dict_index == -1) 
			{
				DEBUGPR(("exit here"));
				return -1;
			}
			curr_dict = word_dist[wdi].dict_index;
			curr_dict_word = word_dist[wdi].start;
			ptr = get_word(&curr_dict_word, dicts[curr_dict]);
			if (!ptr) 
			{
				DEBUGPR(("next dict is empty"));
				return -2;
			}
			else 
				send_to_main("w");
		}
		else 
			send_to_main("w");
		if (dicts[curr_dict]->curr_word > word_dist[wdi].end) 
		{
			curr_dict 	   = word_dist[++wdi].dict_index;
			curr_dict_word = word_dist[wdi].start;
		}

		if (opts.flags & OPT_DICT_COMBO) 
		{
			char *ptr_un, *ptr_pw;
			char buff[MAX_USERNAME_LEN+MAX_PASSWORD_LEN];
			strcpy(buff, ptr);
			ptr = buff;
			if ((ptr_un = strchr(ptr, ' ')) || (ptr_un = strchr(ptr, '\t')))
			{
				memcpy(saved_un, ptr, ptr_un - ptr);
				saved_un[ptr_un - ptr] = 0;
				ptr_pw = ++ptr_un;
				if (!*ptr_pw) 
				{
					dict_errors++;
					if (dict_errors > max_dict_errors)
						fatal("wrong entry in dictionary");
					else
						write_log(log_name, wl_error, "wrong entry in dictionary (%s)", ptr);
					ptr_pw = " ";
				}
				ptr_pw = alltrim(ptr_pw); 
				strcpy(saved_word, ptr_pw);
				ptr = saved_word;
			}
			else
			{
				dict_errors++;
				if (dict_errors > max_dict_errors)
					fatal("wrong entry in dictionary");
				else
					write_log(log_name, wl_error, "wrong entry in dictionary (%s)", ptr);
				strcpy(saved_word, " ");
				ptr = " ";
			}
		}
		else
			strcpy(saved_word, ptr);
		
		if (!(opts.flags & OPT_NOTRANSF)) curr_trans = 1;
	}
	else 
	{
		while (!(ptr = (*trans_flist[curr_trans-1])(saved_word))) 
		{
			if (trans_flist[++curr_trans - 1] == NULL) 
			{
				curr_trans = 0;
				goto get_next_word;
			}
		}
	}

	
	if (opts.flags & OPT_DEF_USER) 
	{
		strcpy(pw, ptr);
		strcpy(un, opts.def_username);

		if (opts.def_username[0]) 
		{
			strcpy(un, opts.def_username);
			strcpy(pw, ptr);
		}
		else 
		{
			strcpy(pw, saved_word);
			strcpy(un, ptr);
		}
	}
	else if (opts.flags & OPT_DEF_PASS) 
	{
		if (opts.def_password[0]) 
		{
			strcpy(pw, opts.def_password);
			strcpy(un, ptr);
		}
		else 
		{
			strcpy(un, saved_word);
			strcpy(pw, ptr);
		}
	}
	else if (opts.flags & OPT_DICT_COMBO)
	{
		strcpy(un, saved_un);
		strcpy(pw, ptr);
	}
	else return -1;
	
	return 0;
}

int
load_dictionaries(dictionary** dlist)
{
	const int mp = 255;
	FILE  *dict_paths;
	char  fbuff[mp];
	int dc;
	char *dlptr;

	if (opts.flags & OPT_DEF_PASSES)
	{
		if (load_dict(opts.def_pass_dict, &def_passes) < 0) 
		{
			fprintf(stderr, "error: couldnt load default passwords\n");
			def_passes = NULL;
		}
	}
	
	if (opts.quick_dict[0])
	{
		if (load_dict(opts.quick_dict, &(dlist[0])) < 0)
		{
			fprintf(stderr, "cannot load dictionary file '%s'\n", fbuff);
			exit(EXIT_FAILURE);
		}
		dlist[1] = NULL;
		totalwords = dlist[0]->word_count;
		return 0;
	}
	
	if (opts.dict_list_file[0])
		dlptr = opts.dict_list_file;
	else
		dlptr = def_dict_paths_file;
	if ((dict_paths = fopen(dlptr, "r")) == NULL) 
	{
		fprintf(stderr, "error: couldnt open %s\n", dlptr);
		return -1;
	}	
	dc = 0;
	while (1) 
	{	
		if (dc > MAX_DICTS) 
		{
			fprintf(stderr, "warning: maximum %d dictionaries allowed\n", MAX_DICTS);
			break;
		}
		if (!fgets(fbuff, mp, dict_paths)) 
		{
			if (feof(dict_paths)) break;
			fprintf(stderr, "error: cant read string from file '%s'\n", dlptr);
			return -2;
		}
		chomp(fbuff);
		DEBUGPR(("loading %s", fbuff));
		fflush(stdout);
		if (load_dict(fbuff, &(dlist[dc])) < 0)
		{
			fprintf(stderr, "dictionary file '%s' skiped\n", fbuff);
		}
		else 
		{
			totalwords += dlist[dc]->word_count;
			dc++;
		}
		dlist[dc+1] = NULL;
	}
	fclose(dict_paths);
	return dc-1;
}

int 
load_dict(char *fn, dictionary **d)
{
	FILE *dict;
	char word[MAX_PASSWORD_LEN];
	
	dict = fopen(fn, "r");
	if (!dict) 
	{
		fprintf(stderr, "warning: cant open '%s'\n", fn);
		return -1;
	}
	
	(*d) = my_alloc(sizeof(dictionary));
	(*d)->wordlist = my_alloc(DICT_BLOCK);
	(*d)->curr_pos   = 0;
	(*d)->word_count = 0;
	(*d)->size		 = 0;
	(*d)->alloc_size = DICT_BLOCK;
	strcpy((*d)->dict_id, "dict_id");
	strcpy((*d)->path, fn);
	
	while (1) 
	{
		if (!fgets(word, MAX_PASSWORD_LEN, dict)) 
		{
			if (feof(dict)) break;
			fprintf(stderr, "error: cant read string from file '%s'\n", fn);
			return -2;
		}
		chomp(word);
		if ((*d)->size + strlen(word) + 1 > (*d)->alloc_size) 
		{
			(*d)->wordlist = my_realloc((*d)->wordlist, 
											 (*d)->alloc_size += DICT_BLOCK);
		}
		(*d)->word_count++;
		strcpy((*d)->wordlist+(*d)->curr_pos, word);
		(*d)->curr_pos += strlen(word) + 1;
		(*d)->size		= (*d)->curr_pos;
	}
	fclose(dict);
	return 0;
}

int
unload_dictionaries(dictionary** dlist)
{
	while (*dlist) 
	{
		free((*dlist)->wordlist);
		free(*dlist++);
	}
	if (def_passes)
	{
		free(def_passes->wordlist);
		free(def_passes);
	}
	return 0;
}

char* 
get_word(int* start_idx, dictionary* d)
{
	char* ptr;
	int wc, sidx;

	if (*start_idx) 
	{
		sidx = *start_idx;
		*start_idx  = 0;
		d->curr_pos = 0;
		wc = 0;
		while (++wc < sidx && (ptr = get_word(start_idx, d)));
		d->curr_word = sidx;
	}
	
	if (d->curr_pos >= d->size) 
	{
		return NULL;
	}
	ptr = d->wordlist + d->curr_pos;
	d->curr_pos += strlen(ptr) + 1;
	d->curr_word++;

	return ptr;
}
